import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import List, Optional
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Console
from rich.style import Style

from vedro._core import (
    Dispatcher,
    ExcInfo,
    Report,
    ScenarioResult,
    StepResult,
    VirtualScenario,
    VirtualStep,
)
from vedro._events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailEvent,
    ScenarioPassEvent,
    ScenarioRunEvent,
    ScenarioSkipEvent,
    StartupEvent,
)
from vedro.plugins.director import RichReporter


@pytest.fixture()
def dispatcher():
    return Dispatcher()


@pytest.fixture()
def console_():
    return Mock(Console)


@pytest.fixture()
def reporter(console_):
    return RichReporter(lambda: console_)


def make_scenario_path(path: str = "", name: str = "scenario.py") -> Path:
    return Path(os.getcwd()) / "scenarios" / path / name


def make_scenario_result(*, scenario_path: Optional[Path] = None,
                         scenario_subject: Optional[str] = None,
                         step_results: Optional[List[StepResult]] = None) -> ScenarioResult:
    scenario_ = Mock(VirtualScenario)
    scenario_.path = scenario_path if scenario_path else make_scenario_path()
    scenario_.subject = scenario_subject

    scenario_result = ScenarioResult(scenario_)
    if step_results:
        for step_result in step_results:
            scenario_result.add_step_result(step_result)
    return scenario_result


def make_step_result(name: Optional[str] = None) -> StepResult:
    step_ = Mock(VirtualStep)
    step_.name = name if step_ else str(step_)
    return StepResult(step_)


def make_exc_info(value: Exception) -> ExcInfo:
    return ExcInfo(type(value), value, None)


@pytest.mark.asyncio
async def test_rich_reporter_arg_parse_event(*, dispatcher: Dispatcher,
                                             reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        parser = ArgumentParser()
        event = ArgParseEvent(parser)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_arg_parsed_event(*, dispatcher: Dispatcher,
                                              reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        args = Namespace(verbose=0)
        event = ArgParsedEvent(args)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_startup_event(*, dispatcher: Dispatcher,
                                           reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = StartupEvent([])

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print("Scenarios")
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event(*, dispatcher: Dispatcher,
                                                reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace = "<namespace>"
        scenario_result = make_scenario_result(scenario_path=make_scenario_path(namespace))
        event = ScenarioRunEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f"* {namespace}", style=Style.parse("bold"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_same_namespace(*, dispatcher: Dispatcher,
                                                               reporter: RichReporter,
                                                               console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace = "<namespace>"
        scenario_result = make_scenario_result(scenario_path=make_scenario_path(namespace))
        event = ScenarioRunEvent(scenario_result)
        await dispatcher.fire(event)
        console_.reset_mock()

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_diff_namespace(*, dispatcher: Dispatcher,
                                                               reporter: RichReporter,
                                                               console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace1 = "<namespace1>"
        scenario_result1 = make_scenario_result(scenario_path=make_scenario_path(namespace1))
        event1 = ScenarioRunEvent(scenario_result1)
        await dispatcher.fire(event1)
        console_.reset_mock()

        namespace2 = "<namespace2>"
        scenario_result2 = make_scenario_result(scenario_path=make_scenario_path(namespace2))
        event2 = ScenarioRunEvent(scenario_result2)

    with when:
        await dispatcher.fire(event2)

    with then:
        assert console_.mock_calls == [
            call.print(f"* {namespace2}", style=Style.parse("bold")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_skip_event(*, dispatcher: Dispatcher,
                                                 reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        scenario_result = make_scenario_result()
        event = ScenarioSkipEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_scenario_pass_event(*, dispatcher: Dispatcher,
                                                 reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        subject = "<subject>"
        scenario_result = make_scenario_result(scenario_subject=subject)
        event = ScenarioPassEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✔ {subject}", style=Style.parse("green"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_default_verbose(*, dispatcher: Dispatcher,
                                                                 reporter: RichReporter,
                                                                 console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        subject = "<subject>"
        scenario_result = make_scenario_result(scenario_subject=subject)
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose0(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(Namespace(verbose=0))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result()
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose1(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(Namespace(verbose=1))
        await dispatcher.fire(event)

        step_name_passed = "<passed step>"
        step_result_passed = make_step_result(step_name_passed).mark_passed()
        step_name_failed = "<failed step>"
        step_result_failed = make_step_result(step_name_failed).mark_failed()
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.print(f"    ✔ {step_name_passed}", style=Style.parse("green")),
            call.print(f"    ✗ {step_name_failed}", style=Style.parse("red")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose2(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(Namespace(verbose=2))
        await dispatcher.fire(event)

        step_result_passed = make_step_result().mark_passed()
        step_result_failed = make_step_result().mark_failed()

        exc_info = make_exc_info(AssertionError())
        step_result_failed.set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.print(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.print(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
            call.print("AssertionError\n", style=Style.parse("yellow")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose3(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(Namespace(verbose=3))
        await dispatcher.fire(event)

        step_result_passed = make_step_result("<passed step>").mark_passed()
        step_result_failed = make_step_result("<failed step>").mark_failed()
        exc_info = make_exc_info(AssertionError())
        step_result_failed.set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        scenario_result.set_scope({"key_int": 1, "key_str": "val"})
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.print(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.print(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
            call.print("AssertionError\n", style=Style.parse("yellow")),
            call.print("Scope:", style=Style.parse("bold blue")),
            call.print(" key_int: ", end="", style=Style.parse("blue")),
            call.print("1"),
            call.print(" key_str: ", end="", style=Style.parse("blue")),
            call.print('"val"'),
            call.print(),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_without_steps_verbose3(*, dispatcher: Dispatcher,
                                                                        reporter: RichReporter,
                                                                        console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(Namespace(verbose=3))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result()
        scenario_result.set_scope({})
        event = ScenarioFailEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_cleanup_event(*, dispatcher: Dispatcher):
    with given:
        console_ = Mock()
        reporter = RichReporter(lambda: console_)
        reporter.subscribe(dispatcher)

        report = Report()
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(),
            call.print("# 0 scenarios, 0 passed, 0 failed, 0 skipped",
                       style=Style.parse("bold red"), end=""),
            call.print(" (0.00s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_success_cleanup_event(*, dispatcher: Dispatcher,
                                                   reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        report = Report()
        scenario_result1 = make_scenario_result().mark_passed()
        scenario_result1.set_started_at(1.0)
        scenario_result1.set_ended_at(2.0)
        report.add_result(scenario_result1)

        scenario_result2 = make_scenario_result().mark_passed()
        scenario_result2.set_started_at(2.0)
        scenario_result2.set_ended_at(3.0)
        report.add_result(scenario_result2)

        scenario_result3 = make_scenario_result().mark_skipped()
        report.add_result(scenario_result3)
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(),
            call.print("# 3 scenarios, 2 passed, 0 failed, 1 skipped",
                       style=Style.parse("bold green"),
                       end=""),
            call.print(" (2.00s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_failed_cleanup_event(*, dispatcher: Dispatcher,
                                                  reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        report = Report()
        scenario_result = make_scenario_result().mark_failed()
        scenario_result.set_started_at(3.145)
        scenario_result.set_ended_at(6.285)
        report.add_result(scenario_result)

        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.print(),
            call.print("# 1 scenarios, 0 passed, 1 failed, 0 skipped",
                       style=Style.parse("bold red"),
                       end=""),
            call.print(" (3.14s)", style=Style.parse("blue"))
        ]



from pathlib import Path
from textwrap import dedent

import pytest

from vedro._core._scenario_loader import ScenarioFileLoader


@pytest.mark.asyncio
async def test_scenario_file_loader(tmp_path: Path):
    path = tmp_path / "scenario.py"
    path.write_text(dedent('''
        import vedro
        class Scenario(vedro.Scenario):
            pass
    '''))

    loader = ScenarioFileLoader()
    scenarios = await loader.load(path)
    assert len(scenarios) == 1


@pytest.mark.asyncio
async def test_template_scenario_file_loader(tmp_path: Path):
    path = tmp_path / "scenario.py"
    path.write_text(dedent('''
        import vedro
        class Scenario(vedro.Scenario):
            @vedro.params("Bob")
            @vedro.params("Alice")
            def __init__(self, user):
                pass
    '''))

    loader = ScenarioFileLoader()
    scenarios = await loader.load(path)
    assert len(scenarios) == 2



import os
from enum import Enum
from typing import Any, Dict, List, Union

from ._step_result import StepResult
from ._virtual_scenario import VirtualScenario

__all__ = ("ScenarioResult", "ScenarioStatus",)


class ScenarioStatus(Enum):
    PENDING = "PENDING"
    PASSED = "PASSED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"


class ScenarioResult:
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario
        self._status: ScenarioStatus = ScenarioStatus.PENDING
        self._started_at: Union[float, None] = None
        self._ended_at: Union[float, None] = None
        self._step_results: List[StepResult] = []
        self._scope: Union[Dict[Any, Any], None] = None

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario

    @property
    def scenario_subject(self) -> Union[str, None]:
        subject = getattr(self._scenario, "subject", None)
        if isinstance(subject, str) and len(subject) > 0:
            return subject.format(**self._scope) if self._scope else subject
        return self._scenario.path.stem.replace("_", " ")

    @property
    def scenario_namespace(self) -> str:
        namespace = os.path.dirname(os.path.relpath(self._scenario.path, "scenarios"))
        return namespace.replace("_", " ").replace("/", " / ")

    def mark_passed(self) -> "ScenarioResult":
        self._status = ScenarioStatus.PASSED
        return self

    def is_passed(self) -> bool:
        return self._status == ScenarioStatus.PASSED

    def mark_failed(self) -> "ScenarioResult":
        self._status = ScenarioStatus.FAILED
        return self

    def is_failed(self) -> bool:
        return self._status == ScenarioStatus.FAILED

    def mark_skipped(self) -> "ScenarioResult":
        self._status = ScenarioStatus.SKIPPED
        return self

    def is_skipped(self) -> bool:
        return self._status == ScenarioStatus.SKIPPED

    @property
    def started_at(self) -> Union[float, None]:
        return self._started_at

    def set_started_at(self, started_at: float) -> None:
        self._started_at = started_at

    @property
    def ended_at(self) -> Union[float, None]:
        return self._ended_at

    def set_ended_at(self, ended_at: float) -> None:
        self._ended_at = ended_at

    def add_step_result(self, step_result: StepResult) -> None:
        self._step_results.append(step_result)

    @property
    def step_results(self) -> List[StepResult]:
        return self._step_results

    def set_scope(self, scope: Dict[Any, Any]) -> None:
        self._scope = scope

    @property
    def scope(self) -> Union[Dict[Any, Any], None]:
        return self._scope

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._scenario!r})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



import os
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from asyncio import CancelledError
from pathlib import Path
from typing import List

from .._events import ArgParsedEvent, ArgParseEvent, CleanupEvent, StartupEvent
from ..plugins import Plugin
from ._dispatcher import Dispatcher
from ._report import Report
from ._runner import Runner
from ._scenario_discoverer import ScenarioDiscoverer

__all__ = ("Lifecycle",)


class Lifecycle:
    def __init__(self, dispatcher: Dispatcher, discoverer: ScenarioDiscoverer) -> None:
        self._dispatcher = dispatcher
        self._discoverer = discoverer
        self._plugins: List[Plugin] = []

    def register_plugins(self, plugins: List[Plugin]) -> None:
        for plugin in plugins:
            self._dispatcher.register(plugin)
            self._plugins.append(plugin)

    async def start(self) -> Report:
        os.chdir(os.path.dirname(os.path.join(os.getcwd(), sys.argv[0])))

        formatter = ArgumentDefaultsHelpFormatter
        arg_parser = ArgumentParser("vedro", formatter_class=formatter, add_help=False)
        await self._dispatcher.fire(ArgParseEvent(arg_parser))
        arg_parser.add_argument("-h", "--help",
                                action="help", help="show this help message and exit")
        args = arg_parser.parse_args()
        await self._dispatcher.fire(ArgParsedEvent(args))

        start_dir = os.path.relpath(Path("scenarios"))
        scenarios = await self._discoverer.discover(Path(start_dir))
        await self._dispatcher.fire(StartupEvent(scenarios))

        runner = Runner(self._dispatcher, (KeyboardInterrupt, SystemExit, CancelledError,))
        report = await runner.run(scenarios)

        await self._dispatcher.fire(CleanupEvent(report))

        return report

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._dispatcher!r}, {self._discoverer!r})"



from pytest import raises

from vedro._core import ScenarioLoader


def test_scenario_loader():
    with raises(Exception) as exc_info:
        ScenarioLoader()

    assert exc_info.type is TypeError
    assert "Can't instantiate abstract class ScenarioLoader" in str(exc_info.value)



from traceback import format_exception
from typing import Callable, Union

from rich.console import Console
from rich.style import Style

from ...._core import Dispatcher
from ...._events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailEvent,
    ScenarioPassEvent,
    ScenarioRunEvent,
    ScenarioSkipEvent,
    StartupEvent,
)
from .._reporter import Reporter
from .utils import format_scope, make_console

__all__ = ("RichReporter",)


class RichReporter(Reporter):
    def __init__(self, console_factory: Callable[[], Console] = make_console) -> None:
        self._console = console_factory()
        self._verbosity = 0
        self._namespace: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioSkipEvent, self.on_scenario_skip) \
                  .listen(ScenarioPassEvent, self.on_scenario_pass) \
                  .listen(ScenarioFailEvent, self.on_scenario_fail) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("-v", "--verbose", action="count", default=self._verbosity)

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._verbosity = event.args.verbose

    def on_startup(self, event: StartupEvent) -> None:
        self._console.print("Scenarios")

    def on_scenario_skip(self, event: ScenarioSkipEvent) -> None:
        pass

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        if event.scenario_result.scenario_namespace != self._namespace:
            self._namespace = event.scenario_result.scenario_namespace
            self._console.print(f"* {self._namespace}", style=Style(bold=True))

    def on_scenario_pass(self, event: ScenarioPassEvent) -> None:
        subject = event.scenario_result.scenario_subject
        self._console.print(f" ✔ {subject}", style=Style(color="green"))

    def on_scenario_fail(self, event: ScenarioFailEvent) -> None:
        subject = event.scenario_result.scenario_subject
        self._console.print(f" ✗ {subject}", style=Style(color="red"))

        if self._verbosity == 0:
            return

        for step_result in event.scenario_result.step_results:
            if step_result.is_passed():
                self._console.print(f"    ✔ {step_result.step_name}",
                                    style=Style(color="green"))
            elif step_result.is_failed():
                self._console.print(f"    ✗ {step_result.step_name}",
                                    style=Style(color="red"))

        if self._verbosity == 1:
            return

        for step_result in event.scenario_result.step_results:
            if step_result.exc_info:
                tb = format_exception(
                    step_result.exc_info.type,
                    step_result.exc_info.value,
                    step_result.exc_info.traceback,
                )
                self._console.print("".join(tb), style=Style(color="yellow"))

        if self._verbosity == 2:
            return

        if event.scenario_result.scope:
            self._console.print("Scope:", style=Style(color="blue", bold=True))
            for key, val in format_scope(event.scenario_result.scope):
                self._console.print(f" {key}: ", end="", style=Style(color="blue"))
                self._console.print(val)
            self._console.print()

    def on_cleanup(self, event: CleanupEvent) -> None:
        if event.report.failed == 0 and event.report.passed > 0:
            style = Style(color="green", bold=True)
        else:
            style = Style(color="red", bold=True)
        self._console.print()
        self._console.print(f"# {event.report.total} scenarios, "
                            f"{event.report.passed} passed, "
                            f"{event.report.failed} failed, "
                            f"{event.report.skipped} skipped",
                            style=style,
                            end="")
        self._console.print(f" ({event.report.elapsed:.2f}s)", style=Style(color="blue"))



import importlib
import importlib.util
import inspect
import os
from pathlib import Path
from typing import List, Type, cast

from ..._scenario import Scenario
from ._scenario_loader import ScenarioLoader

__all__ = ("ScenarioFileLoader",)


class ScenarioFileLoader(ScenarioLoader):
    def _path_to_module_name(self, path: Path) -> str:
        return ".".join(path.with_suffix("").parts)

    async def load(self, path: Path) -> List[Type[Scenario]]:
        module_name = self._path_to_module_name(path)

        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None:
            # ImportError?
            raise Exception(path)
        module = importlib.util.module_from_spec(spec)

        cast(importlib.abc.Loader, spec.loader).exec_module(module)

        loaded = []
        for name in dir(module):
            val = getattr(module, name)
            if inspect.isclass(val) and issubclass(val, Scenario):
                if not val.__name__.startswith("Scenario"):
                    continue
                val.__file__ = os.path.abspath(module.__file__)
                loaded.append(val)
        return loaded
